package pwc.taxtech.atms.vat.service.impl;

import org.apache.poi.ss.formula.OperationEvaluationContext;
import org.apache.poi.ss.formula.eval.NumberEval;
import org.apache.poi.ss.formula.eval.StringEval;
import org.apache.poi.ss.formula.eval.ValueEval;
import org.apache.poi.ss.formula.functions.FreeRefFunction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import pwc.taxtech.atms.common.datasource.ShardingContextHolder;
import pwc.taxtech.atms.common.util.MyAsserts;
import pwc.taxtech.atms.common.util.SpringContextUtil;
import pwc.taxtech.atms.dto.vatdto.BBParasBo;
import pwc.taxtech.atms.dto.vatdto.CellTemplatePerGroupDto;
import pwc.taxtech.atms.dto.vatdto.CloseableFormulaDataSource;
import pwc.taxtech.atms.dto.vatdto.CurrentPeriodBo;
import pwc.taxtech.atms.dto.vatdto.ReportCellDataSourceDto;
import pwc.taxtech.atms.exception.Exceptions;
import pwc.taxtech.atms.vat.entity.CellData;
import pwc.taxtech.atms.vat.entity.Report;
import pwc.taxtech.atms.vat.service.impl.report.functions.FormulaContext;

import java.math.BigDecimal;

public class ReportFormulaFactory {
    static final Logger LOGGER = LoggerFactory.getLogger(ReportFormulaFactory.class);
    final static ValueEval defaultEval = new StringEval("0");
    final static BigDecimal defaultBigDecimal = new BigDecimal("0");

    public static abstract class ReportBaseFormula implements FreeRefFunction {
        final FormulaContext context;
        final FormulaAgent agent;

        protected ReportBaseFormula(FormulaContext context) {
            this.context = context;
            this.agent = context.getFormulaAgent();
        }

        public int getYear(int parameterYear) {
            if (parameterYear <= -1) {
                return parameterYear + context.getYear();
            } else if (parameterYear == 0) {
                return context.getYear();
            } else {
                return parameterYear;
            }
        }

        @Override
        public ValueEval evaluate(ValueEval[] args, OperationEvaluationContext ec) {
            try {
                return wrapExceptionEval(args, ec);
            } catch (Exception e) {
                LOGGER.warn("un expect exception", e);
                e.printStackTrace();
                return defaultEval;
            }
        }

        public abstract ValueEval wrapExceptionEval(ValueEval[] args, OperationEvaluationContext ec) throws Exception;

    }

    public static class BB extends ReportBaseFormula {
        protected BB(FormulaContext context) {
            super(context);
        }

        @Override
        public ValueEval wrapExceptionEval(ValueEval[] args, OperationEvaluationContext ec) throws Exception {
            return new NumberEval(bb(new BBParasBo(args, ec)).doubleValue());
        }

        public BigDecimal bb(BBParasBo bo) throws Exception {
            try (CloseableFormulaDataSource closeDataSource = new CloseableFormulaDataSource(context.getFormulaAgent())) {
                CurrentPeriodBo curPeriod = CurrentPeriodBo.getPeriod(bo.getPeriod().intValue(), context.getPeriod());
                curPeriod.fixedCurYear(getYear(curPeriod.getCurYear()));

                ReportCellDataSourceDto nullCellDto = ReportCellDataSourceDto.nullDataSource(bo, curPeriod);
                closeDataSource.addDS(nullCellDto);

                CellTemplatePerGroupDto cellTemplateData = agent.getCellTemplateGroupDto(context.getReportTemplateGroupID(),
                        context.getProjectID()).stream().filter(dto -> dto.getRowIndex() == bo.getColumnIndex() - 1
                        && dto.getColumnIndex() == bo.getColumnIndex() - 1).findFirst().get();


                MyAsserts.assertNotNull(cellTemplateData, Exceptions.BB_CELL_TEMP_NULL);

                nullCellDto.fixedWithGroup(cellTemplateData);

                // todo: fix datasource name by templateList(neo)
                if (curPeriod.getCurPeriod() == -99) {
                    closeDataSource.clean();
                    BigDecimal returnEval = defaultBigDecimal;
                    if (context.getPeriod() <= 1) {
                        return defaultBigDecimal;
                    }

                    for (int p = 1; p < context.getPeriod(); p++) {
                        returnEval = returnEval.add(bb(new BBParasBo(bo, p, curPeriod.getCurYear())));
                    }

                    return returnEval;
                }

                bo.disCount();
                String dbName = SpringContextUtil.formulaAdminMapper.getPastProjectDbName(curPeriod.getCurYear(), context.getOrganizationID());
                String currentProjectDb = ShardingContextHolder.getDataSourceKey();
                ShardingContextHolder.setDataSourceKey(dbName);

                try {
                    Report report = agent.getReportByTemplate(cellTemplateData.getReportTemplateID(), curPeriod.getCurPeriod());
                    MyAsserts.assertNotNull(report, Exceptions.BB_REPORT_NULL);

                    CellData cellData = agent.getCellDataListByTemplate(cellTemplateData.getCellTemplateID(), report.getId());
                    MyAsserts.assertNotNull(cellData, Exceptions.BB_CELL_DATA_NULL);

                    nullCellDto = ReportCellDataSourceDto.extractFromGroup(bo, curPeriod, cellData, cellTemplateData);
                    // todo: fix datasource name by templateList(neo)

                    MyAsserts.assertNotNull(cellData.getData(), Exceptions.BB_CELL_DATA_NULL);

//                  cellValue=  RoundValue(cellValue, cellDataType)TODO:maybe fixd round by cellDataTyep(KV neo)
                    BigDecimal cellValue = new BigDecimal(cellData.getData()).setScale(4,
                            BigDecimal.ROUND_HALF_DOWN);
                    nullCellDto.setAmount(cellValue);
                    LOGGER.debug("cell static value ");
                    ShardingContextHolder.setDataSourceKey(currentProjectDb);
                    return cellValue;
                } catch (Exception e) {
                    LOGGER.warn("un support exception", e);
                    ShardingContextHolder.setDataSourceKey(currentProjectDb);
                    throw e;
                }
            }
        }


    }

}
